from abc import ABC, abstractmethod
from pathlib import Path
from dataclasses import dataclass
from typing import Union
from .info import Info, load_info

@dataclass(frozen=True)
class FileRule:
    prop_name: str
    file_name: str
    required: bool = True


class BaseDirectoryStructure(ABC):
    def __init__(self):
        self.rules = self.get_rules()

        for rule in self.rules:
            setattr(self, rule.prop_name, rule.file_name)

    @abstractmethod
    def get_rules(self) -> list[FileRule]:
        """Return the list of file rules required for this folder structure."""
        pass

    def validate(self, folder: Path) -> None:
        missing = [
            rule.file_name
            for rule in self.rules
            if rule.required and not (folder / rule.file_name).is_file()
        ]
        if missing:
            raise FileNotFoundError(f"Missing required files: {', '.join(missing)}")


class SITLDirectoryStructure(BaseDirectoryStructure):
    def get_rules(self) -> list[FileRule]:
        return [
            FileRule("info_file", "info.toml"),
            FileRule("params_file", "params.airframe"),
            FileRule("modules_file", "sitl.modules"),
            FileRule("params_post_file", "params.airframe.post", required=False),
        ]



class FirmwareDirectoryStructure(BaseDirectoryStructure):
    def get_rules(self) -> list[FileRule]:
        return [
            FileRule("info_file", "info.toml"),
            FileRule("params_file", "params.airframe"),
            FileRule("modules_file", "board.modules"),
            FileRule("params_post_file", "params.airframe.post", required=False),
        ]

def valid_dir_path(path: Union[str, Path]) -> Path:

    directory = None

    if isinstance(path, str):
        directory = Path(path)
    elif isinstance(path, Path):
        directory = path
    else:
        raise TypeError(f"path must be str or Path, got {type(path).__name__}")

    if not directory.exists():
        raise FileNotFoundError(f"Path does not exist: {directory}")
    if not directory.is_dir():
        raise NotADirectoryError(f"Expected a directory but got: {directory}")

    return directory

class Directory:
    def __init__(self, path: Union[str, Path], dir_type: str):

        self.directory = valid_dir_path(path)

        self.dir_type = dir_type

        self.__structure = self.__validate_structure()

        for rule in self.__structure.get_rules():
            setattr(self, rule.prop_name, rule.file_name)

        self.__info_manager = load_info(self.directory / self.__structure.info_file)
        self.__info = self.__info_manager.get_info()

    def __validate_structure(self):
        structure = None
        if self.dir_type == "sitl":
            structure = SITLDirectoryStructure()
        elif self.dir_type == "firmware":
            structure = FirmwareDirectoryStructure()
        else:
            raise ValueError(f"Unknown build type: {self.dir_type}")

        structure.validate(self.directory)
        return structure

    def get_info(self) -> Info:
        return self.__info

    @property
    def info(self) -> dict:
        return self.__info_manager.info


def load_directory(path: Union[str, Path], dir_type: str) -> Directory:

    return Directory(path, dir_type)



import sys
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

import pytest
import easy_px4_utils

def test_load_info_good():

    info = easy_px4_utils.load_info("""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "v1.15.4"
        custom_fw_version = "0.0.2"
    """)

    info = easy_px4_utils.load_info("""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "v1.15.4"
        px4_commit = "main"
        custom_fw_version = "0.0.2"
    """)

    info_test_file = Path(__file__).resolve().parent.parent / "demos/protoflyer/info.toml"
    easy_px4_utils.load_info(info_test_file)
    easy_px4_utils.load_info(str(info_test_file))

def test_load_info_missing_key():

    with pytest.raises(KeyError):
        easy_px4_utils.load_info("""
            name = "drone"
            id = 12345
            vendor = "px4"
            px4_version = "v1.15.4"
            custom_fw_version = "0.0.2"
        """)


def test_load_info_single_component():
    info = easy_px4_utils.load_info("""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "v1.15.4"
        custom_fw_version = "0.0.2"
        components = "some_component"
    """)

def test_load_info_list_components():
    info = easy_px4_utils.load_info("""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "v1.15.4"
        custom_fw_version = "0.0.2"
        components = ["some_component", "other_component"]
    """)

def test_load_info_bad():

    with pytest.raises(tomllib.TOMLDecodeError):
        # incorrect formating name
        easy_px4_utils.load_info("""
            name = drone
            id = "12345"
            vendor = "px4"
            model = "fmu-v3"
            px4_version = "v1.15.4"
            custom_fw_version = "0.0.2"
        """)


    with pytest.raises(TypeError):
        easy_px4_utils.load_info("""
            name = 'drone'
            id = '12345'
            vendor = "px4"
            model = "fmu-v3"
            px4_version = "v1.15.4"
            custom_fw_version = "0.0.2"
        """)

    with pytest.raises(tomllib.TOMLDecodeError):
        # incorrect formating vendor
        easy_px4_utils.load_info("""
            name = drone
            id = '12345'
            vendor = px4
            model = "fmu-v3"
            px4_version = "v1.15.4"
            custom_fw_version = "0.0.2"
        """)

    with pytest.raises(TypeError):
        # extra keys that is not part of the info structure
        easy_px4_utils.load_info("""
            name = "drone"
            id = 12345
            vendor = "px4"
            model = "fmu-v3"
            px4_version = "v1.15.4"
            custom_fw_version = "0.0.2"
            extra = "stuff"
        """)


valid_px4_versions = [
    "v1.15.4",
    "v1.15.4-beta1",
    "v1.15.4-alpha3",
    "v1.15.4-rc22",
    "v1.15.4-dev",
]

@pytest.mark.parametrize("version", valid_px4_versions)
def test_valid_px4_versions(version):
    easy_px4_utils.load_info(f"""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "{version}"
        custom_fw_version = "1.2.3"
    """)

invalid_px4_versions = [
    "v1.15.4.0",
    "v1.15.4-beta",
    "v1.15.4-rc",
    "v1.15.4-alpha",
    "v1.15.4-dev1",
    "1.14",
    "latest",
    "1.15.4",
]

@pytest.mark.parametrize("version", invalid_px4_versions)
def test_invalid_px4_versions(version):
    with pytest.raises(ValueError):
        easy_px4_utils.load_info(f"""
            name = "drone"
            id = 12345
            vendor = "px4"
            model = "fmu-v3"
            px4_version = "{version}"
            custom_fw_version = "0.0.2"
        """)

valid_custom_fw_versions = [
    "1.2.3",
    "1.2.3-rc2",
    "1.2.3-alpha2",
    "1.2.3-beta2",
    "1.2.3-dev",
]

@pytest.mark.parametrize("version", valid_custom_fw_versions)
def test_valid_custom_fw_versions(version):
    easy_px4_utils.load_info(f"""
        name = "drone"
        id = 12345
        vendor = "px4"
        model = "fmu-v3"
        px4_version = "v1.15.4"
        custom_fw_version = "{version}"
    """)


invalid_custom_fw_versions = [
    "0.2.0.0",
    "0.2",
]

@pytest.mark.parametrize("version", invalid_custom_fw_versions)
def test_invalid_custom_fw_versions(version):
    with pytest.raises(ValueError):
        easy_px4_utils.load_info(f"""
            name = "drone"
            id = 12345
            vendor = "px4"
            model = "fmu-v3"
            px4_version = "v1.15.4"
            custom_fw_version = "{version}"
        """)



import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

import re
from pathlib import Path
from typing import Union, Optional, get_origin, get_args
from argparse import ArgumentTypeError
from dataclasses import dataclass

@dataclass
class Info:
    """
    Class for keeping track structure of the info file.
    You can conceptually think of this class as the spec for the info.toml file.
    """
    name: str
    id: int
    vendor: str
    model: str
    px4_version: str
    px4_commit: Optional[str] = None

    custom_fw_version: Optional[str] = "0.0.0"
    components: Optional[Union[str, list[str]]] = None


class InfoManager:

    def __init__(self, input_info: Union[str, Path]) -> None:

        self.__content: str | None = None

        if isinstance(input_info, str):
            possible_path = Path(input_info)
            if possible_path.is_file():
                self.path = possible_path
            else:
                self.path = None
                self.__content = input_info
        elif isinstance(input_info, Path):
            if input_info.is_file():
                self.path = input_info
            else:
                raise FileNotFoundError(f"{input_info} is not a file.")

        else:
            raise TypeError(f"path must be str or Path, got {type(input_info).__name__}")

        self.__info_dict_raw: dict = self.__parse()
        self.__info = self.__load_dict(self.__info_dict_raw)


    def __parse(self) -> dict:
        try:
            if self.__content is not None:
                info = tomllib.loads(self.__content)
            elif self.path is not None:
                with self.path.open("r", encoding="utf-8") as f:
                    content = f.read()
                info = tomllib.loads(content)
            else:
                raise ValueError("No content or file path to parse from.")
        except tomllib.TOMLDecodeError as e:
            raise tomllib.TOMLDecodeError(f"Problem parsing TOML.\nTraceback:\n\n{e}")
        except Exception as e:
            raise Exception(f"Unexpected error: {e}")

        return info


    def __matches_type(self, value, expected_type) -> bool:
        """
        Return True if `value` conforms to `expected_type`.  Works for:
          - basic built-ins, e.g. str, int, float
          - Optional[...]  (i.e. Union[..., NoneType])
          - Union[...]     (checks any one branch)
          - list[SomeType]
          - dict[KeyType,ValueType]
          - (and will treat other non‐generic types as simple isinstance checks)
        """

        origin = get_origin(expected_type)
        args = get_args(expected_type)

        # 1) If it’s a Union (including Optional), try each branch:
        if origin is Union:
            # “Union[..., NoneType]” is the same as Optional[...]
            for branch in args:
                # Allow None if branch is NoneType
                if branch is type(None) and value is None:
                    return True
                # Otherwise, recursively check the other branch
                if branch is not type(None) and self.__matches_type(value, branch):
                    return True
            return False

        # 2) If it’s a list[T], first check isinstance(value, list), then each element matches T
        if origin is list:
            if not isinstance(value, list):
                return False
            (item_type,) = args
            for element in value:
                if not self.__matches_type(element, item_type):
                    return False
            return True

        # 3) If it’s a dict[K, V], check isinstance(value, dict) and each key/value pair matches
        if origin is dict:
            if not isinstance(value, dict):
                return False
            key_type, val_type = args
            for k, v in value.items():
                if not self.__matches_type(k, key_type) or not self.__matches_type(v, val_type):
                    return False
            return True

        return isinstance(value, expected_type)

    def __validation_types(self, info_dict) -> bool:

        for expected_key, expected_type in Info.__annotations__.items():
            if expected_key not in info_dict:
                # maybe is Optional
                if get_origin(expected_type) is Union and type(None) in get_args(expected_type):
                    continue
                else:
                    raise KeyError(f"{self.path}: Missing required field: {expected_key} of type {type(expected_type)}")

            value = info_dict[expected_key]

            if value is None:
                # If the field really is Optional[...] (Union[..., NoneType]), None is okay.
                if get_origin(expected_type) is Union and type(None) in get_args(expected_type):
                    continue
                else:
                    raise TypeError(f"Field '{expected_key}' must be of type {expected_type}, but got None.")

            if not self.__matches_type(value, expected_type):
                raise TypeError(
                    f"Field '{expected_key}' must be of type '{expected_type.__name__}'. "
                    f"Got value={value!r} (type={type(value).__name__})"
                )

        return True


    def __validation_content(self, info_dict) -> bool:

        if not re.fullmatch(r'v([0-9]+)\.([0-9]+)\.[0-9]+((-dev)|(-alpha[0-9]+)|(-beta[0-9]+)|(-rc[0-9]+))?$', info_dict["px4_version"]):
            raise ValueError(f"'px4_version' must be in format v<major>.<minor>.<patch>[-rc<rc>|-beta<beta>|-alpha<alpha>|-dev]. Got {info_dict['px4_version']}")

        if info_dict.get("custom_fw_version") is not None:
            if not re.fullmatch(r'([0-9]+)\.([0-9]+)\.([0-9]+)((-dev)|(-alpha[0-9]+)|(-beta[0-9]+)|(-?rc[0-9]+))?$', info_dict['custom_fw_version']):
                raise ValueError(f"'custom_fw_version' must be semantic version <major>.<minor>.<patch>[-rc<rc>|-beta<beta>|-alpha<alpha>|-dev]. Got {info_dict['custom_fw_version']}")

        return True


    def __load_dict(self, info_dict: dict) -> Info:

        self.__validation_types(info_dict)

        self.__validation_content(info_dict)

        return Info(**info_dict)

    def get_info(self) -> Info:
        return self.__info

    @property
    def info(self) -> dict:
        return self.__info_dict_raw

def load_info(info: Union[str, Path]) -> InfoManager:
    """
    Load info.toml file from path.
    """
    return InfoManager(info)

def load_info_dict(info: Union[str, Path]) -> dict[str, dict]:
    """
    Load info.toml file from path.
    """
    return InfoManager(info).info
